# SPDX-License-Identifier: MIT
from __future__ import annotations

from dataclasses import dataclass, field

from ..tokens import (
    KeyBlockToken,
    KeyMultiToken,
    KeyValueToken,
    Mark,
    SequenceBlockToken,
    SequenceMultiToken,
    SequenceValueToken,
)
from ..types import TokenGenerator, YAMLCollection, YAMLMapping, YAMLSequence


@dataclass
class Tokenizer:
    o: YAMLCollection

    line: int = 1

    indents: list[int] = field(default_factory=list)

    @property
    def column(self) -> int:
        return sum(self.indents) + 1

    @property
    def mark(self) -> Mark:
        return Mark(line=self.line, column=self.column)

    def tokenize(self) -> TokenGenerator:
        value = self.o
        match value:
            case dict():
                yield from self.handle_mapping(value)
            case list():
                yield from self.handle_sequence(value)
            case _:
                raise ValueError(f"{value}, {type(value)}")

    def handle_mapping(self, o: YAMLMapping) -> TokenGenerator:
        for key, value in o.items():
            match value:
                case None:
                    yield KeyBlockToken(key=key, mark=self.mark)
                    self.line += 1
                case dict():
                    yield KeyBlockToken(key=key, mark=self.mark)
                    self.line += 1
                    self.indents.append(4)
                    yield from self.handle_mapping(value)
                    self.indents.pop()
                case list():
                    yield KeyBlockToken(key=key, mark=self.mark)
                    self.line += 1
                    self.indents.append(4)
                    yield from self.handle_sequence(value)
                    self.indents.pop()
                case str():
                    key_mark = self.mark
                    lines = value.splitlines()
                    if len(lines) == 1:
                        yield KeyValueToken(key=key, value=value, mark=key_mark)
                    else:
                        self.indents.append(4)
                        self.line += 1
                        yield KeyMultiToken(
                            key=key,
                            value=value,
                            mark=key_mark,
                            value_mark=self.mark,
                        )
                        self.indents.pop()
                    self.line += len(lines)
                case _:
                    raise ValueError(f"{value}, {type(value)}")

    def handle_sequence(self, o: YAMLSequence) -> TokenGenerator:
        for value in o:
            match value:
                case dict():
                    yield SequenceBlockToken(mark=self.mark)
                    self.indents.append(2)
                    yield from self.handle_mapping(value)
                    self.indents.pop()
                case str():
                    sequence_mark = self.mark
                    lines = value.splitlines()
                    if len(lines) == 1:
                        yield SequenceValueToken(value=value, mark=sequence_mark)
                    else:
                        self.indents.append(4)
                        self.line += 1
                        yield SequenceMultiToken(
                            value=value, mark=sequence_mark, value_mark=self.mark
                        )
                        self.indents.pop()
                    self.line += len(lines)
                case _:
                    raise ValueError(f"{value}, {type(value)}")


def tokenize(o: YAMLCollection) -> TokenGenerator:
    yield from Tokenizer(o).tokenize()



# SPDX-License-Identifier: MIT
from ...tests import catalog
from ..parser import parse
from ..scanner import scan


def test_parser() -> None:
    with catalog.se.open(encoding="utf-8") as file:
        result = parse(scan(file))
    assert result



# SPDX-License-Identifier: MIT
from .__main__ import (
    from_stream,
    from_string,
    to_stream,
    to_string,
)

__all__ = [
    "from_stream",
    "from_string",
    "to_stream",
    "to_string",
]



# SPDX-License-Identifier: MIT
from io import StringIO

from ..__main__ import from_stream, to_stream
from ..types import YAMLMapping
from . import catalog

catalog_dict: YAMLMapping = {}


def test_from_stream() -> None:
    with catalog.se.open(encoding="utf-8") as file:
        result = from_stream(file, expect=dict)

    global catalog_dict
    catalog_dict = result
    assert result


def test_to_stream() -> None:
    stream = StringIO()

    to_stream(stream, catalog_dict)



# SPDX-License-Identifier: MIT
from __future__ import annotations

from dataclasses import dataclass
from typing import IO

from ..constants import CHAR, PATTERN, SYMBOL
from ..tokens import Mark
from .results import LiteralResult


@dataclass
class Reader:
    stream: IO[str]
    done: bool = False

    buffer: str = ""
    line: int = 0
    column: int = 0

    @property
    def indent(self) -> int:
        return self.column - 1

    @property
    def mark(self) -> Mark:
        return Mark(line=self.line, column=self.column)

    def readline(self) -> bool:
        self.buffer = self.stream.readline()

        self.line += 1
        self.column = 1

        if len(self.buffer) == 0:
            return False
        else:
            return True

    def advance(self, index: int = 1) -> None:
        self.column += index
        self.buffer = self.buffer[index:]

    def chomp_whitespace(self) -> None:
        before = len(self.buffer)
        self.buffer = self.buffer.lstrip()
        self.column += before - len(self.buffer)

    def get_literal(self) -> tuple[str, Mark, LiteralResult]:
        self.chomp_whitespace()
        mark = self.mark
        buffer = self.buffer

        if buffer.startswith(SYMBOL.COMMENT):
            return "", mark, LiteralResult.COMMENT

        if buffer.startswith(PATTERN.MULTI):
            return "", mark, LiteralResult.MULTI

        quoted: str
        if buffer.startswith(CHAR.QUOTE):
            quoted = buffer[0]
            buffer = buffer[1:]
        else:
            quoted = ""

        reason: LiteralResult
        chunks: str = ""
        skipped: int = 0
        while buffer:
            if buffer.startswith(CHAR.LINEBREAK):
                skipped += 1
                reason = LiteralResult.VALUE
                break
            elif quoted and buffer.startswith(quoted):
                skipped += 1
                reason = LiteralResult.VALUE
                break
            elif buffer.startswith(PATTERN.KEY):
                skipped += len(PATTERN.KEY)
                reason = LiteralResult.KEY
                break
            elif buffer.startswith(PATTERN.BLOCK):
                skipped += len(PATTERN.BLOCK)
                reason = LiteralResult.BLOCK
                break
            elif buffer.startswith(PATTERN.COMMENT):
                skipped += len(PATTERN.COMMENT)
                reason = LiteralResult.COMMENT
                break
            else:
                chunks += buffer[0]
                buffer = buffer[1:]
                continue
        else:
            reason = LiteralResult.VALUE

        self.advance(len(chunks) + skipped)
        return chunks, mark, reason



# SPDX-License-Identifier: MIT
from enum import Enum, auto


class ReaderResult(Enum):
    CONTINUE = auto()
    END = auto()


class LiteralResult(Enum):
    KEY = auto()
    VALUE = auto()
    COMMENT = auto()
    BLOCK = auto()
    MULTI = auto()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Optional

from .tokens import Mark


@dataclass
class MarkedError(Exception):
    problem: str
    mark: Optional[Mark]

    def __str__(self) -> str:
        return f"{self.problem}\nOccured at: `{self.mark}`"



# SPDX-License-Identifier: MIT
from ...tests import fullschema
from ..parser import parse
from ..scanner import scan


def test_scanner() -> None:
    with fullschema.se.open(encoding="utf-8") as file:
        result = list(scan(file))

    expected = fullschema.tokens
    assert result == expected


def test_parser() -> None:
    with fullschema.se.open(encoding="utf-8") as file:
        result = parse(scan(file))

    expected = fullschema.de
    assert result == expected



# SPDX-License-Identifier: MIT
from typing import IO
from ..types import TokenGenerator
from ..constants import CHAR, PATTERN
from ..tokens import (
    KeyBlockToken,
    KeyMultiToken,
    KeyValueToken,
    SequenceBlockToken,
    SequenceMultiToken,
    SequenceValueToken,
)


def serialize(stream: IO[str], tokens: TokenGenerator) -> None:
    for token in tokens:
        stream.write(token.indent * CHAR.SPACE)
        match token:
            case KeyValueToken():
                stream.write(token.key + PATTERN.KEY + token.value + CHAR.LF)
            case KeyMultiToken():
                stream.write(token.key + PATTERN.KEY + PATTERN.MULTI)

                for line in token.value.splitlines():
                    if len(line) == 0:
                        stream.write(CHAR.LF)
                    else:
                        stream.write(token.value_indent * CHAR.SPACE + line + CHAR.LF)
            case KeyBlockToken():
                stream.write(token.key + PATTERN.BLOCK)
            case SequenceValueToken():
                stream.write(PATTERN.SEQUENCE + token.value + CHAR.LF)
            case SequenceMultiToken():
                stream.write(PATTERN.SEQUENCE + PATTERN.MULTI)

                for line in token.value.splitlines():
                    if len(line) == 0:
                        stream.write(CHAR.LF)
                    else:
                        stream.write(token.value_indent * CHAR.SPACE + line + CHAR.LF)
            case SequenceBlockToken():
                stream.write(PATTERN.SEQUENCE)

                token = next(tokens)
                match token:
                    case KeyValueToken():
                        stream.write(token.key + PATTERN.KEY + token.value + CHAR.LF)
                    case KeyBlockToken():
                        stream.write(token.key + PATTERN.BLOCK)
                    case _:
                        raise TypeError(f"Expected `{KeyValueToken | KeyBlockToken}` but got `{type(token)}`")
            case _:
                raise NotImplementedError(f"Missing implementation for {type(token)}")



# SPDX-License-Identifier: MIT
from pathlib import Path

from ..tokens import Mark, SequenceMultiToken, SequenceValueToken
from ..types import YAMLSequence

tokens = [
    SequenceValueToken(mark=Mark(line=1, column=1), value="One"),
    SequenceValueToken(mark=Mark(line=2, column=1), value="Two"),
    SequenceValueToken(mark=Mark(line=3, column=1), value="Three"),
    SequenceValueToken(mark=Mark(line=4, column=1), value="Four"),
    SequenceMultiToken(mark=Mark(line=5, column=1), value="Five.A\n\nFive.B", value_mark=Mark(line=6, column=5)),
]

de: YAMLSequence = ["One", "Two", "Three", "Four", "Five.A\n\nFive.B"]

se = Path(__file__).parent.joinpath("sequence.yaml")



# SPDX-License-Identifier: MIT
from pathlib import Path

from ..tokens import KeyMultiToken, Mark
from ..types import YAMLMapping

tokens = [
    KeyMultiToken(
        key="multiline_string",
        value="This is a multiline paragraph\nAnd this is the second paragraph.\n\nAnd a third paragraph.",
        mark=Mark(line=1, column=1),
        value_mark=Mark(line=2, column=5),
    ),
]

de: YAMLMapping = {
    "multiline_string": "This is a multiline paragraph\nAnd this is the second paragraph.\n\nAnd a third paragraph.",
}

se = Path(__file__).parent.joinpath("multiline.yaml")



# SPDX-License-Identifier: MIT
from __future__ import annotations
from typing import IO, Optional

from ..types import TokenGenerator
from ..errors import MarkedError
from .reader import Reader
from .results import LiteralResult
from ..tokens import (
    KeyBlockToken,
    KeyMultiToken,
    KeyValueToken,
    SequenceBlockToken,
    SequenceMultiToken,
    SequenceValueToken,
)
from ..constants import CHAR, PATTERN, SYMBOL


class ScannerError(MarkedError):
    pass


class Scanner:
    scanner: Scanner

    def __init__(self, stream: IO[str]):
        self.reader = Reader(stream)

    def scan(self) -> TokenGenerator:
        multi: Optional[KeyMultiToken | SequenceMultiToken] = None

        while True:
            if not self.reader.readline():
                if multi is not None:
                    multi.value = multi.value.rstrip(CHAR.LF)
                    yield multi
                break

            # Indent
            self.reader.chomp_whitespace()

            if multi is not None:
                # NOTE: Check if multline is initialized
                if multi.value_mark is None:
                    multi.value_mark = self.reader.mark

                if self.reader.indent < multi.value_indent and len(self.reader.buffer) > 0:
                    multi.value = multi.value.rstrip(CHAR.LF)
                    yield multi
                    multi = None
                else:
                    value, _, result = self.reader.get_literal()
                    multi.value += value + CHAR.LF
                    continue

            # Parse
            buffer = self.reader.buffer
            if len(self.reader.buffer) == 0:
                pass
            elif buffer.startswith(SYMBOL.COMMENT):
                # NOTE: No need to consume since we proceed to next line
                pass
            elif buffer.startswith(CHAR.LINEBREAK):
                # NOTE: No need to consume since we proceed to next line
                value, value_mark, result = self.reader.get_literal()
            elif buffer.startswith(PATTERN.SEQUENCE):
                sequence_mark = self.reader.mark
                self.reader.advance(len(PATTERN.SEQUENCE))

                value, value_mark, result = self.reader.get_literal()
                match result:
                    case LiteralResult.VALUE | LiteralResult.COMMENT:
                        yield SequenceValueToken(value, mark=sequence_mark)
                    case LiteralResult.MULTI:
                        multi = SequenceMultiToken("", mark=sequence_mark, value_mark=None)
                    case LiteralResult.BLOCK:
                        yield SequenceBlockToken(mark=sequence_mark)
                        yield KeyBlockToken(value, mark=value_mark)
                    case LiteralResult.KEY:
                        yield SequenceBlockToken(mark=sequence_mark)

                        key, key_mark = value, value_mark
                        value, value_mark, result = self.reader.get_literal()
                        match result:
                            case LiteralResult.VALUE | LiteralResult.COMMENT:
                                yield KeyValueToken(key, value, mark=key_mark)
                            case LiteralResult.KEY | LiteralResult.BLOCK:
                                raise ScannerError("Expected a value, got a key.", value_mark)
                            case LiteralResult.MULTI:
                                multi = KeyMultiToken(key, "", mark=key_mark, value_mark=None)
            else:
                key, key_mark, result = self.reader.get_literal()
                match result:
                    case LiteralResult.KEY:
                        value, mark, result = self.reader.get_literal()
                        match result:
                            case LiteralResult.VALUE | LiteralResult.COMMENT:
                                yield KeyValueToken(key, value, mark=key_mark)
                            case LiteralResult.MULTI:
                                multi = KeyMultiToken(key, "", mark=key_mark, value_mark=None)
                            case LiteralResult.KEY:
                                raise ScannerError("Nested mappings are not allowed on the same line.", mark)
                            case LiteralResult.BLOCK:
                                raise ScannerError("Nested mappings are not allowed on the same line.", mark)
                    case LiteralResult.BLOCK:
                        yield KeyBlockToken(key, mark=key_mark)
                    case LiteralResult.COMMENT:
                        raise ScannerError("Expected a key, got a comment instead", key_mark)
                    case LiteralResult.VALUE:
                        raise ScannerError("Expected a key, got a value.", key_mark)
                    case LiteralResult.MULTI:
                        raise ScannerError("Expected a key, got a multiline literal.", key_mark)


def scan(stream: IO[str]) -> TokenGenerator:
    return Scanner(stream).scan()



# SPDX-License-Identifier: MIT
from pathlib import Path

from ..tokens import KeyValueToken, Mark
from ..types import YAMLMapping

tokens = [
    KeyValueToken(key="A", value="B", mark=Mark(line=1, column=1)),
]

de: YAMLMapping = {"A": "B"}

se = Path(__file__).parent.joinpath("singleline.yaml")



# SPDX-License-Identifier: MIT
from typing import Generator

from .tokens import Token

type YAMLMapping = dict[str, YAMLObject]
type YAMLSequence = list[YAMLObject]
type YAMLCollection = YAMLMapping | YAMLSequence
type YAMLObject = YAMLCollection | str | None
type TokenGenerator = Generator[Token, None, None]



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from io import StringIO
from typing import IO

from .de.parser import ParserError, parse
from .de.scanner import ScannerError, scan
from .errors import MarkedError
from .se.serializer import serialize
from .se.tokenizer import tokenize
from .types import YAMLCollection


@dataclass
class DeserializerError(MarkedError):
    stream: IO[str]

    def __str__(self) -> str:
        return str(self.stream) + "\n" + super().__str__()


def from_stream[T: YAMLCollection](stream: IO[str], *, expect: type[T] = dict):
    try:
        data = parse(scan(stream))
        if not isinstance(data, expect):
            raise TypeError(f"Expected {expect.__name__}, got {type(data).__name__}")
        return data
    except (ParserError, ScannerError) as e:
        raise DeserializerError(e.problem, e.mark, stream)
    except Exception as e:
        print(stream)
        raise e


def from_string[T: YAMLCollection](source: str, expect: type[T]):
    stream = StringIO(source)
    return from_stream(stream, expect=expect)


def to_stream(stream: IO[str], data: YAMLCollection) -> None:
    serialize(stream, tokenize(data))


def to_string(data: YAMLCollection) -> str:
    stream = StringIO()
    to_stream(stream, data)
    return stream.getvalue()



# SPDX-License-Identifier: MIT
from pathlib import Path

from ..tokens import KeyBlockToken, KeyValueToken, Mark
from ..types import YAMLMapping

tokens = [
    KeyValueToken(key="name", value="Kazuo Shinohara", mark=Mark(line=1, column=1)),
    KeyValueToken(key="based", value="japan.tokyo", mark=Mark(line=2, column=1)),
    KeyBlockToken(key="empty", mark=Mark(line=3, column=1)),
    KeyBlockToken(key="projects", mark=Mark(line=4, column=1)),
    KeyBlockToken(key="house-of-white", mark=Mark(line=5, column=5)),
    KeyValueToken(key="year", value="1966", mark=Mark(line=6, column=9)),
    KeyValueToken(key="born", value="1925", mark=Mark(line=7, column=1)),
]

de: YAMLMapping = {
    "name": "Kazuo Shinohara",
    "based": "japan.tokyo",
    "empty": None,
    "projects": {
        "house-of-white": {
            "year": "1966",
        },
    },
    "born": "1925",
}

se = Path(__file__).parent.joinpath("mapping.yaml")



# SPDX-License-Identifier: MIT
from enum import EnumType, StrEnum


class SYMBOL(StrEnum):
    KEY = ":"
    SEQUENCE = "-"
    COMMENT = "#"
    MULTI = "|"


class CHAR(EnumType):
    """https://docs.python.org/3/library/stdtypes.html#str.splitlines"""

    SPACE = " "
    TAB = "\t"
    WHITESPACE = (SPACE, TAB)
    LF = "\n"
    CR = "\r"
    CRLF = "\r\n"
    FF = "\f"
    FS = "\x1c"
    GS = "\x1d"
    RS = "\x1e"
    NL = "\x85"
    LS = "\u2028"
    PS = "\u2028"
    LINEBREAK = (LF, CR, CRLF, FF, FS, GS, RS, NL, LS, PS)
    NULL = "\0"
    END = (NULL, *LINEBREAK)
    SINGLEQUOTE = "'"
    DOUBLEQUOTE = '"'
    QUOTE = (SINGLEQUOTE, DOUBLEQUOTE)
    RESERVED = tuple((value.value for value in SYMBOL.__members__.values()))


class PATTERN(StrEnum):
    SEQUENCE = f"{SYMBOL.SEQUENCE}{CHAR.SPACE}"
    COMMENT = f"{CHAR.SPACE}{SYMBOL.COMMENT}"
    KEY = f"{SYMBOL.KEY}{CHAR.SPACE}"
    BLOCK = f"{SYMBOL.KEY}{CHAR.LF}"
    MULTI = f"{SYMBOL.MULTI}{CHAR.LF}"



# SPDX-License-Identifier: MIT
from pathlib import Path

from ..tokens import (
    KeyBlockToken,
    KeyMultiToken,
    KeyValueToken,
    Mark,
    SequenceBlockToken,
    SequenceValueToken,
)
from ..types import YAMLMapping

tokens = [
    KeyValueToken(mark=Mark(line=1, column=1), key="simple_map", value="SimpleKey"),
    KeyBlockToken(mark=Mark(line=2, column=1), key="nested_map"),
    KeyValueToken(mark=Mark(line=3, column=5), key="number", value="123"),
    KeyValueToken(mark=Mark(line=4, column=5), key="squoted", value="Quote"),
    KeyValueToken(mark=Mark(line=5, column=5), key="dquoted", value="QuoteQuote"),
    KeyBlockToken(mark=Mark(line=6, column=1), key="empty_key"),
    KeyBlockToken(mark=Mark(line=7, column=1), key="sequence"),
    SequenceValueToken(mark=Mark(line=8, column=5), value="Item1"),
    SequenceValueToken(mark=Mark(line=9, column=5), value="Item2"),
    KeyBlockToken(mark=Mark(line=10, column=1), key="seq_in_map"),
    KeyBlockToken(mark=Mark(line=11, column=5), key="seq1"),
    SequenceValueToken(mark=Mark(line=12, column=9), value="One"),
    SequenceValueToken(mark=Mark(line=13, column=9), value="Two"),
    KeyBlockToken(mark=Mark(line=14, column=5), key="seq2"),
    SequenceValueToken(mark=Mark(line=15, column=9), value="OneOne"),
    SequenceValueToken(mark=Mark(line=16, column=9), value="TwoTwo"),
    KeyBlockToken(mark=Mark(line=17, column=1), key="map_in_seq"),
    SequenceBlockToken(mark=Mark(line=18, column=5)),
    KeyBlockToken(mark=Mark(line=18, column=7), key="map1A"),
    SequenceValueToken(mark=Mark(line=19, column=11), value="A"),
    SequenceValueToken(mark=Mark(line=20, column=11), value="AA"),
    KeyValueToken(mark=Mark(line=21, column=7), key="map1B", value="B"),
    SequenceBlockToken(mark=Mark(line=22, column=5)),
    KeyBlockToken(mark=Mark(line=22, column=7), key="map2"),
    SequenceValueToken(mark=Mark(line=23, column=11), value="1"),
    SequenceValueToken(mark=Mark(line=24, column=11), value="2"),
    KeyBlockToken(mark=Mark(line=25, column=1), key="simple_in_seq"),
    SequenceBlockToken(mark=Mark(line=26, column=5)),
    KeyValueToken(mark=Mark(line=26, column=7), key="One", value="1"),
    SequenceBlockToken(mark=Mark(line=27, column=5)),
    KeyValueToken(mark=Mark(line=27, column=7), key="Two", value="2"),
    KeyMultiToken(
        mark=Mark(line=28, column=1),
        key="multiline_string",
        value="This is a multiline paragraph\nAnd this is the second paragraph.",
        value_mark=Mark(line=29, column=5),
    ),
    KeyBlockToken(mark=Mark(line=31, column=1), key="closing_key"),
]

de: YAMLMapping = {
    "simple_map": "SimpleKey",
    "nested_map": {
        "number": "123",
        "squoted": "Quote",
        "dquoted": "QuoteQuote",
    },
    "empty_key": None,
    "sequence": [
        "Item1",
        "Item2",
    ],
    "seq_in_map": {
        "seq1": [
            "One",
            "Two",
        ],
        "seq2": [
            "OneOne",
            "TwoTwo",
        ],
    },
    "map_in_seq": [
        {
            "map1A": [
                "A",
                "AA",
            ],
            "map1B": "B",
        },
        {
            "map2": [
                "1",
                "2",
            ],
        },
    ],
    "simple_in_seq": [
        {"One": "1"},
        {"Two": "2"},
    ],
    "multiline_string": "This is a multiline paragraph\nAnd this is the second paragraph.",
    "closing_key": None,
}

se = Path(__file__).parent.joinpath("fullschema.yaml")



# SPDX-License-Identifier: MIT
from pathlib import Path

from ..tokens import (
    KeyBlockToken,
    KeyValueToken,
    Mark,
    SequenceBlockToken,
    SequenceValueToken,
)
from ..types import YAMLSequence

tokens = [
    SequenceBlockToken(mark=Mark(line=1, column=1)),
    KeyBlockToken(key="map1A", mark=Mark(line=1, column=3)),
    SequenceValueToken(value="A", mark=Mark(line=2, column=7)),
    SequenceValueToken(value="AA", mark=Mark(line=3, column=7)),
    KeyValueToken(key="map1B", value="B", mark=Mark(line=4, column=3)),
    KeyBlockToken(key="map1C", mark=Mark(line=5, column=3)),
    SequenceBlockToken(mark=Mark(line=6, column=1)),
    KeyBlockToken(key="map2", mark=Mark(line=6, column=3)),
    SequenceValueToken(value="1", mark=Mark(line=7, column=7)),
    SequenceValueToken(value="2", mark=Mark(line=8, column=7)),
]

de: YAMLSequence = [
    {
        "map1A": [
            "A",
            "AA",
        ],
        "map1B": "B",
        "map1C": None,
    },
    {
        "map2": [
            "1",
            "2",
        ],
    },
]

se = Path(__file__).parent.joinpath("sequence_mapping.yaml")



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any

from ..errors import MarkedError
from ..tokens import (
    KeyBlockToken,
    KeyMultiToken,
    KeyValueToken,
    SequenceBlockToken,
    SequenceMultiToken,
    SequenceValueToken,
)
from ..types import TokenGenerator, YAMLCollection


class ParserError(MarkedError):
    pass


@dataclass
class Parser:
    tokens: TokenGenerator
    collections: dict[int, YAMLCollection] = field(default_factory=dict)

    def parse(self) -> YAMLCollection:
        try:
            first = next(self.tokens)
        except StopIteration:
            return {}

        if first.indent != 0:
            raise ParserError(
                f"First token's indentation is `{first.indent}` and not `0`", first.mark
            )

        match first:
            case KeyValueToken() | KeyMultiToken():
                mapping: dict[Any, Any] = {}
                mapping[first.key] = first.value
                self.collections[first.indent] = mapping
            case KeyBlockToken():
                mapping = {}
                mapping[first.key] = {}
                self.collections[first.indent] = mapping
            case SequenceValueToken() | SequenceMultiToken():
                sequence: list[Any] = []
                sequence.append(first.value)
                self.collections[first.indent] = sequence
            case SequenceBlockToken():
                sequence = []
                self.collections[first.indent] = sequence
            case _:
                raise NotImplementedError(f"Missing implementation for {type(first)}")

        for token in self.tokens:
            parent: YAMLCollection
            match token:
                case KeyValueToken() | KeyMultiToken():
                    parent = self.get_collection(token.indent, dict)
                    parent[token.key] = token.value
                case KeyBlockToken():
                    parent = self.get_collection(token.indent, dict)
                    parent[token.key] = None
                case SequenceValueToken() | SequenceMultiToken():
                    parent = self.get_collection(token.indent, list)
                    parent.append(token.value)
                case SequenceBlockToken():
                    _parent = self.get_collection(token.indent, list)
                case _:
                    raise NotImplementedError(
                        f"Missing implementation for {type(token)}"
                    )

        self.close_collections(first.indent)
        return self.collections[first.indent]

    def get_collection[T: YAMLCollection](self, target_indent: int, typ: type[T]):
        self.close_collections(target_indent)

        if target_indent in self.collections:
            collection = self.collections[target_indent]

            if collection is not None and not isinstance(collection, typ):
                raise TypeError(f"Expected `{typ}`, but got `{type(collection)}`")
        else:
            collection = typ()
            self.collections[target_indent] = collection

        return collection  # type: ignore

    def close_collections(self, target_indent: int) -> None:
        indents: list[int] = sorted(self.collections.keys())
        while indents:
            indent: int = indents.pop()
            if indent > target_indent:
                collection = self.collections.pop(indent)
                parent = self.collections[indents[-1]]

                match parent:
                    case dict():
                        parent[list(parent.keys())[-1]] = collection
                    case list():
                        if not isinstance(collection, dict):
                            raise TypeError(
                                f"Expected `{dict}`, but got `{type(collection)}`"
                            )
                        else:
                            parent.append(collection)
                    case _:
                        raise TypeError(
                            f"Expected {list} or {dict}, but got `{type(parent)}`"
                        )
            else:
                break


def parse(tokens: TokenGenerator) -> YAMLCollection:
    return Parser(tokens).parse()



# SPDX-License-Identifier: MIT
from __future__ import annotations

from dataclasses import dataclass, field
from types import NoneType
from typing import Optional


@dataclass(kw_only=True)
class Mark:
    line: int
    column: int

    def __repr__(self) -> str:
        return f"Line: {self.line}, Column: {self.column}"


@dataclass(kw_only=True)
class TokenMixin:
    mark: Mark

    @property
    def indent(self) -> int:
        return self.mark.column - 1


@dataclass()
class KeyMixin(TokenMixin):
    key: str


@dataclass()
class ValueMixin(TokenMixin):
    value: str
    value_mark: Optional[Mark] = field(default_factory=NoneType)

    @property
    def value_indent(self) -> int:
        return self.value_mark.column - 1  # type: ignore


@dataclass
class KeyValueToken(ValueMixin, KeyMixin):
    pass


@dataclass
class KeyMultiToken(ValueMixin, KeyMixin):
    pass


@dataclass
class KeyBlockToken(KeyMixin):
    pass


@dataclass
class SequenceValueToken(ValueMixin):
    pass


@dataclass
class SequenceBlockToken(TokenMixin):
    pass


@dataclass
class SequenceMultiToken(ValueMixin):
    pass


type Token = (
    KeyValueToken | KeyMultiToken | KeyBlockToken | SequenceValueToken | SequenceMultiToken | SequenceBlockToken
)



# SPDX-License-Identifier: MIT
from io import StringIO

from ...tests import mapping
from ..serializer import serialize
from ..tokenizer import tokenize


def test_tokenizer() -> None:
    result = list(tokenize(mapping.de))

    expected = mapping.tokens
    assert result == expected


def test_serializer() -> None:
    stream = StringIO()
    serialize(stream, tokenize(mapping.de))
    result = stream.getvalue()

    expected = mapping.se.read_text(encoding="utf-8")
    assert result == expected
